!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : common_mod
!> @brief   REIN common modules
!! @author  Naoyuki Miyashita (NM)
!! @date    NM: 2009/04/02
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module common_mod

  use constants_mod
  use mpi

  implicit none
  private

  ! REIN input variables
  character(icha_2h), public  :: Ginput_replica_rank
  character(icha_2h), public  :: Ginput_replica_values
  character(icha_2h), public  :: Ginput_replica_on
  character(icha_2h), public  :: Ginput_exchange_pattern
  character(icha_2h), public  :: Ginput_axis_exchange
  character(icha_2h), public  :: Ginput_restart_tscale

  character(icha_2h), public  :: Goutput_restart_rank
  character(icha_2h), public  :: Goutput_restart_tscale
  character(icha_2h), public  :: Goutput_replica_rank
  character(icha_2h), public  :: Goutput_replica_energy

  character(icha_2h), public  :: Gprogram_name
  character(icha_2h), public  :: Gparallel_exec
  character(icha_2h), public  :: Gparallel_module_md
  character(icha_2h), public  :: Goption_md
  character(icha_2h), public  :: Gextend_option_md
  character(icha_2h), public  :: Gvelocity_rescaling
  double precision,   public  :: Gbase_temperature

  integer,            public  :: Gfirst_step
  integer,            public  :: Gequiv_steps
  integer,            public  :: Gproduct_steps

  character(icha_2h), public  :: Gparameter_setting ! on or off, default is off only for equiv
  integer,            public  :: Gindex_of_initial_parameter ! only work for equiv

  character(icha_2h), public  :: Ginitial_file_name
  character(icha_2h), public  :: Gio_file_name

  integer,            public  :: Gnumber_of_axis
  integer,            public  :: Gnumber_of_exchange
  integer,            public  :: Gaxis_index(imax_axis)
  character(icha_2h), public  :: Gaxis_type_name(imax_axis)
  integer,            public  :: Gnumber_of_replicas(imax_axis)
  integer,            public  :: Gaxis_control_param(imax_axis)
  integer,            public  :: Gaxis_num_of_param(imax_axis)
  integer,            public  :: Gaxis_index_coef(imax_param, imax_axis)
  character(icha_2h), public  :: Gperiodic(1:imax_axis)
  double precision,   public  :: Gperiodic_values(imax_axis, 2)

  integer,            public  :: Gsleep_time
  integer,            public  :: Gnumber_of_cpus_for_exec
    

  ! communicator information
  type, public:: s_comm_x
    integer                         :: comm
    integer                         :: rank
    integer                         :: nprocs
  end type s_comm_x

  ! main information
  type, public :: s_main_info
    integer                         :: total_replicas
    integer                         :: num_of_values
    integer                         :: flag_velocity_rescaling
    integer                         :: flag_has_harmonic
    integer                         :: exchange_axis_switch

    integer,            allocatable :: replica_rank(:,:)
    double precision,   allocatable :: replica_values(:,:)
    double precision,   allocatable :: replica_values_tmp(:,:,:)
    integer,            allocatable :: replica_on(:,:)
    integer,            allocatable :: replica_ep(:,:)
    integer,            allocatable :: axis_exchange_index(:)
    double precision,   allocatable :: replica_tscale(:,:)
    integer,            allocatable :: replica_no(:,:)

    double precision,   allocatable :: replica_energy(:)
    double precision,   allocatable :: replica_potentiale(:)
    double precision,   allocatable :: replica_distance(:)
    double precision,   allocatable :: replica_extra_pote(:)
    double precision,   allocatable :: replica_temperature(:)
    double precision,   allocatable :: replica_rep_temp(:)
    double precision,   allocatable :: replica_kinetice(:)
    integer,            allocatable :: replica_prerank(:,:)
    integer,            allocatable :: replica_param2rep(:,:)
    double precision,   allocatable :: replica_param_p(:,:)
    integer,            allocatable :: exchange_switch(:)

    integer,            allocatable :: master_nrank_array(:,:)

    integer                         :: initial_steps
    integer                         :: last_steps
    integer                         :: i_steps
    integer                         :: i_equiv
    integer                         :: i_product
    integer,            allocatable :: exchange_success(:)
    integer                         :: exchange_total
    double precision,   allocatable :: exchange_ratio(:)

    double precision                :: clockt
    double precision                :: clockt2

  end type s_main_info
  
  ! local information
  type, public :: s_local_info

    integer                         :: nrank                    = 0

    character(icha_2h)              :: base_file                = ''
    character(icha_2h)              :: local_dir                = ''
    character(icha_2h)              :: local_file               = ''
    character(icha_2h)              :: local_path               = ''
    character(icha_2h)              :: local_input_file         = ''
    character(icha_2h)              :: local_input_path         = ''
    character(icha_2h)              :: local_output_file        = ''
    character(icha_2h)              :: local_output_path        = ''
    character(icha_2h)              :: local_pre_input_file     = ''
    character(icha_2h)              :: local_pre_input_path     = ''
    character(icha_2h)              :: local_command_file       = ''
    character(icha_2h)              :: local_command_path       = ''
                                    
    character(icha_2h)              :: local_extra_file1        = ''
    character(icha_2h)              :: local_extra_path1        = ''
    character(icha_2h)              :: local_extra_file2        = ''
    character(icha_2h)              :: local_extra_path2        = ''
    character(icha_2h)              :: local_extra_file3        = ''
    character(icha_2h)              :: local_extra_path3        = ''
                                    
    character(icha_2h)              :: local_input_crd          = ''
    character(icha_2h)              :: local_output_crd         = ''
    character(icha_2h)              :: local_output_trj         = ''
    character(icha_2h)              :: local_output_prop        = ''
                                    
    integer                         :: local_nline_inputfile    = 0
    integer                         :: local_nline_crdfile      = 0
    integer                         :: local_nline_extrafile1   = 0
    integer                         :: local_nbyte_velfile      = 0
                                    
    character(icha_2h)              :: local_command            = ''
                                    
    character(icha_8 )              :: name_pref_rank           = ''
    character(icha_8 )              :: name_i_steps             = ''
    character(icha_8 )              :: name_ipre_steps          = ''
                                    
    integer                         :: num_axis_type            = 0
                                    
    double precision,   allocatable :: condition_values(:)

  end type s_local_info

  ! subroutines
  public  :: init_values
  public  :: bcast_values

  public  :: trans_int2cha
  public  :: trans_cha2int
  public  :: random_numb
  public  :: norm_random
  public  :: get_mpi_loop
  public  :: periodic_inbox
  private :: n_each_core_max
  private :: i_local_order

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  init_values
  !> @brief      set initial values to variables
  !! @authors    NT, NM
  !! @date       2012/06/23 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine init_values

    Ginput_replica_rank     = "file_rank.d"
    Ginput_replica_values   = "file_value.d"
    Ginput_replica_on       = "file_on.d"
    Ginput_exchange_pattern = "file_ex.d"
    Ginput_axis_exchange    = "file_axis_ex.d"
    Ginput_restart_tscale   = "file_coef.d"

    Goutput_restart_rank    = "file_rank.d"
    Goutput_restart_tscale  = "file_coef.d"
    Goutput_replica_rank    = "replica.trj"
    Goutput_replica_energy  = "replica_energy.trj"

    Gprogram_name           = "namd"
    Gparallel_exec          = " ./namd2 "
    Gparallel_module_md     = ""
    Goption_md              = ""
    Gextend_option_md       = ""
    Gvelocity_rescaling     = "v_namd_bin" ! t-rescaling rearrang
                                           ! "v_namd_bin"
                                           ! "v_namd"
                                           ! "namd"
                                           ! "v_marble"
                                           ! "marble"
    Gbase_temperature       = 300.d0

    Gfirst_step             = 1
    Gequiv_steps            = 0
    Gproduct_steps          = 10
    
    Gparameter_setting      = "off"
    Gindex_of_initial_parameter= 1

    Ginitial_file_name      = "initial"
    Gio_file_name           = "rep."

    Gnumber_of_axis         = 1
    Gnumber_of_exchange     = 2
    Gnumber_of_replicas(1:imax_axis)   = 0

    Gperiodic(1)            = "off"
    Gperiodic_values(1,1)   = -180
    Gperiodic_values(1,2)   =  180
    Gperiodic_values(2,1)   = -180
    Gperiodic_values(2,2)   =  180
    Gperiodic(2)            = "off"

    Gsleep_time             = 1
    Gnumber_of_cpus_for_exec= 1

    return

  end subroutine init_values


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  bcast_values
  !> @brief      broadcast values to other rank variables
  !! @authors    NT, NM
  !! @date       2009/11/20 (NM)
  !! @param[in]  world MPI information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine bcast_values(world)

    ! formal arguments
    type(s_comm_x), intent(in) :: world

    ! local variables
    integer :: ierr

    
    call mpi_bcast(Ginput_replica_rank, icha_2h,     &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Ginput_replica_values, icha_2h,   &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Ginput_replica_on, icha_2h,       &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Ginput_exchange_pattern, icha_2h, &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Ginput_axis_exchange, icha_2h,    &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Ginput_restart_tscale, icha_2h,   &
         MPI_CHARACTER, 0, world%comm, ierr)

    call mpi_bcast(Goutput_restart_rank, icha_2h,    &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Goutput_restart_tscale, icha_2h,  &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Goutput_replica_rank, icha_2h,    &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Goutput_replica_energy, icha_2h,  &
         MPI_CHARACTER, 0, world%comm, ierr)

    call mpi_bcast(Gprogram_name, icha_2h,           &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Gparallel_exec, icha_2h,          &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Gparallel_module_md, icha_2h,     &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Goption_md, icha_2h,              &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Gextend_option_md, icha_2h,       &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Gvelocity_rescaling, icha_2h,     &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Gbase_temperature, 1, &
         MPI_DOUBLE_PRECISION, 0, world%comm, ierr)

    call mpi_bcast(Gfirst_step, 1,                   &
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gequiv_steps, 1,                  &
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gproduct_steps, 1,                &
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gparameter_setting, icha_2h,      &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Gindex_of_initial_parameter, 1,   &
         MPI_INTEGER, 0, world%comm, ierr)

    call mpi_bcast(Ginitial_file_name, icha_2h,      &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Gio_file_name, icha_2h,           &
         MPI_CHARACTER, 0, world%comm, ierr)

    call mpi_bcast(Gnumber_of_axis, 1,               &
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gnumber_of_exchange, 1,           &
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gaxis_index(1), imax_axis,        &
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gaxis_type_name(1), icha_2h*imax_axis, &
         MPI_CHARACTER, 0, world%comm, ierr)
    call mpi_bcast(Gnumber_of_replicas(1), imax_axis,&
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gaxis_control_param(1), imax_axis,&
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gaxis_num_of_param(1), imax_axis, &
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gaxis_index_coef(1,1), imax_axis*imax_param,&
         MPI_INTEGER, 0, world%comm, ierr)
    call mpi_bcast(Gperiodic(1), icha_2h*imax_axis,  &
         MPI_CHARACTER, 0, world%comm, ierr)      
    call mpi_bcast(Gperiodic_values(1,1), imax_axis*2, &
         MPI_DOUBLE_PRECISION, 0, world%comm, ierr)

    call mpi_bcast(Gsleep_time, 1,                   &
         MPI_INTEGER, 0 , world%comm, ierr)
    call mpi_bcast(Gnumber_of_cpus_for_exec,1,       &
         MPI_INTEGER, 0, world%comm, ierr)

    return

  end subroutine bcast_values


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  trans_int2cha
  !> @brief      translate integer to character
  !! @authors    NT, NM
  !! @date       2009/06/26 
  !! @param[in]  i  integer number
  !! @param[out] ni character
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine trans_int2cha(i, ni)

    ! formal arguments
    integer,         intent(in)    :: i
    character(8),    intent(out)   :: ni
    
    
    select case(i)
    case(0:9)               ; write(ni,'(I1)') i
    case(10:99)             ; write(ni,'(I2)') i
    case(100:999)           ; write(ni,'(I3)') i
    case(1000:9999)         ; write(ni,'(I4)') i
    case(10000:99999)       ; write(ni,'(I5)') i
    case(100000:999999)     ; write(ni,'(I6)') i
    case(1000000:9999999)   ; write(ni,'(I7)') i
    case(10000000:99999999) ; write(ni,'(I8)') i
    end select

    return

  end subroutine trans_int2cha


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  trans_cha2int
  !> @brief      translate character to integer
  !! @authors    NT, NM
  !! @date       2009/06/26 
  !! @param[in]  ni character
  !! @param[out] i  integer number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine trans_cha2int(ni, i)

    ! formal arguments
    character(8),    intent(in)    :: ni
    integer,         intent(out)   :: i


    read(ni,*) i

    return

  end subroutine trans_cha2int


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  random_numb
  !> @brief      generate random number
  !! @authors    NT, NM
  !! @date       2009/06/26
  !! @param[inout] d  random number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine random_numb(d)

    ! formal arugments
    double precision, intent(inout) :: d

    ! local variables
    integer           :: v(1:8), c
#ifdef GFORT
    integer,allocatable :: igseed(:)
    integer             :: nseed,i
#endif


    call date_and_time(values=v)
    call system_clock(count=c)

    c = c+int(v(8)*10000+v(7)*1000+v(6)*100+v(5)*10+v(4))+int(d*10000)

#ifdef GFORT
    call random_seed(size=nseed)
    allocate(igseed(nseed))
    igseed = c + 37*(/(i-1,i=1,nseed,1)/)
    call random_seed(put=igseed)
    deallocate(igseed)
#else
    call random_seed(put=(/c/))
#endif

    call random_number(d)

    return

  end subroutine random_numb
  

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  hi_random_numb
  !> @brief      generate random number
  !! @authors    NT, NM
  !! @date       2001/06/26
  !! @param[inout] d  random number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine hi_random_numb(d)
    
    ! parameter
    double precision, parameter     :: Limit = 0.001d0

    ! formal arguments
    double precision, intent(inout) :: d


    if (d < Limit) then 
      d = d*1000
    end if

    call random_number(d)

    return

  end subroutine hi_random_numb


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  norm_random
  !> @brief      generate random number
  !! @authors    NT, NM
  !! @date       2009/06/26 
  !! @param[out]   ans
  !! @param[inout] f
  !! @param[in]    mu
  !! @param[in]    sigma
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine norm_random(ans, f, mu, sigma)

    ! forma arguments
    double precision, intent(out)   :: ans
    double precision, intent(inout) :: f
    double precision, intent(in)    :: mu
    double precision, intent(in)    :: sigma

    ! local variables
    double precision  :: head, sin_value


    call random_number(f)

    head = -2.d0*dlog(f)
    head = dsqrt(head)

    call random_number(f)

    sin_value = dsin(2.d0 * PI * f)

    ans = mu + sigma * (head * sin_value)

    return

  end subroutine norm_random
  

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  get_mpi_loop
  !> @brief      get the mpi loop
  !! @authors    NT, NM
  !! @date       2012/02/26
  !! @param[in]    nreplica  number of replicas
  !! @param[in]    nprocs    number of process
  !! @param[inout] mpi_loop  number of ranks
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_mpi_loop(nreplica, nprocs, mpi_loop)

    ! formal arguments
    integer, intent(in)    :: nprocs
    integer, intent(in)    :: nreplica
    integer, intent(inout) :: mpi_loop(1:nprocs,1:3)

    ! local variables
    integer  :: bin, ini, added
    integer  :: i


    ini = 1
    added = 0

    do i = 0, nprocs - 1
      bin = n_each_core_max(i, nprocs, nreplica)
      mpi_loop(i+1,1) = ini
      mpi_loop(i+1,2) = ini + bin -1
      mpi_loop(i+1,3) = bin
      ini = ini + bin
    end do
    !
    return

  end subroutine get_mpi_loop
  

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  periodic_inbox
  !> @brief      periodic inbox
  !! @authors    NT, NM
  !! @date       2012/03/26 
  !! @param[inout]  x
  !! @param[in]     min
  !! @param[in]     max
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine periodic_inbox(x, min, max)

    ! formal arguments
    double precision, intent(inout) :: x
    double precision, intent(in)    :: min
    double precision, intent(in)    :: max

    ! local variables
    double precision :: diff


    diff = max - min
    if (x <= min) &
      x = x + diff
    if (x >= max) &
      x = x - diff

    return

  end subroutine periodic_inbox
  

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function  n_each_core_max
  !> @brief      
  !! @authors  NT, NM
  !! @date     2012/06/26
  !! @param[in] i_rank
  !! @param[in] n_rank
  !! @param[in] n_max
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  integer function n_each_core_max(i_rank, n_rank, n_max)

    ! formal arguments
    integer, intent(in) :: i_rank
    integer, intent(in) :: n_rank
    integer, intent(in) :: n_max

    ! local variables
    integer  :: a, b


    a = i_local_order(n_max, n_rank)
    b = int(n_max/n_rank)
    n_each_core_max = b

    if (i_rank+1 <= a .and. a /= n_rank ) &
      n_each_core_max = n_each_core_max + 1

    return

  end function n_each_core_max


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function  i_local_order
  !> @brief    local numbering  1,2,3,4, 1,2,3,4, 1,2,3,4
  !! @authors  NT, NM
  !! @date     2012/06/26 
  !! @param[in]  i
  !! @param[in]  n_max
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  integer function i_local_order(i, n_max)

    ! formal arguments
    integer, intent(in) :: i
    integer, intent(in) :: n_max

    ! local variables
    integer :: a


    a = mod(i, n_max)
    if (a == 0) &
      a = n_max
    i_local_order = a

    return

  end function i_local_order

end module common_mod
